@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2008-2012 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2008-2018 -- leonerd@leonerd.org.uk
 
 package IO::Async::Loop::Epoll;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.15';
+our $VERSION = '0.18';
 use constant API_VERSION => '0.49';
 
 # Only Linux is known always to be able to report EOF conditions on
@@ -24,10 +24,15 @@ use Carp;
 use Linux::Epoll 0.005;
 
 use POSIX qw( EINTR EPERM SIG_BLOCK SIG_UNBLOCK sigprocmask sigaction ceil );
+use Scalar::Util qw( refaddr );
 
 use constant _CAN_WATCHDOG => 1;
 use constant WATCHDOG_ENABLE => IO::Async::Loop->WATCHDOG_ENABLE;
 
+use Struct::Dumb;
+
+struct SignalWatch => [qw( code pending orig )];
+
 =head1 NAME
 
 C<IO::Async::Loop::Epoll> - use C<IO::Async> with C<epoll> on Linux
@@ -67,7 +72,7 @@ This subclass of L<IO::Async::Loop> uses C<epoll(7)> on Linux to perform
 read-ready and write-ready tests so that the OZ<>(1) high-performance
 multiplexing of Linux's C<epoll_pwait(2)> syscall can be used.
 
-The C<epoll> Linux subsystem uses a persistant registration system, meaning
+The C<epoll> Linux subsystem uses a persistent registration system, meaning
 that better performance can be achieved in programs using a large number of
 filehandles. Each C<epoll_pwait(2)> syscall only has an overhead proportional
 to the number of ready filehandles, rather than the total number being
@@ -85,7 +90,9 @@ handle file IO and signals concurrently.
 
 =cut
 
-=head2 $loop = IO::Async::Loop::Epoll->new()
+=head2 new
+
+   $loop = IO::Async::Loop::Epoll->new()
 
 This function returns a new instance of a C<IO::Async::Loop::Epoll> object.
 
@@ -107,9 +114,14 @@ sub new
 
    $self->{fakeevents} = {};
 
-   $self->{signals} = {};
+   $self->{signals} = {}; # {$name} => SignalWatch
    $self->{masks} = {};
 
+   # epoll gets very upset if applications close() filehandles without telling
+   # it, and then try to add that mask a second time. We can attempt to detect
+   # this by storing the mapping from fileno to refaddr($fh)
+   $self->{refaddr_for_fileno} = {};
+
    return $self;
 }
 
@@ -137,7 +149,9 @@ sub DESTROY
    }
 }
 
-=head2 $count = $loop->loop_once( $timeout )
+=head2 loop_once
+
+   $count = $loop->loop_once( $timeout )
 
 This method calls C<epoll_pwait(2)>, and processes the results of that call.
 It returns the total number of C<IO::Async::Notifier> callbacks invoked, or
@@ -156,6 +170,8 @@ sub loop_once
    # Round up to next milisecond to avoid zero timeouts
    my $msec = defined $timeout ? ceil( $timeout * 1000 ) : -1;
 
+   $timeout = 0 if keys %{ $self->{fakeevents} };
+
    my $ret = $self->{epoll}->wait( $self->{maxevents}, $msec / 1000, $self->{sigmask} );
 
    return undef if !defined $ret and $! != EINTR;
@@ -195,9 +211,9 @@ sub loop_once
 
    my $signals = $self->{signals};
    foreach my $sigslot ( values %$signals ) {
-      if( $sigslot->[1] ) {
-         $sigslot->[0]->();
-         $sigslot->[1] = 0;
+      if( $sigslot->pending ) {
+         $sigslot->pending = 0;
+         $sigslot->code->();
          $count++;
       }
    }
@@ -264,7 +280,11 @@ sub watch_io
    my $fakeevents = $self->{fakeevents};
 
    if( !$curmask ) {
-      if( eval { $epoll->add( $handle, \@bits, $cb ); 1 } ) {
+      defined $self->{refaddr_for_fileno}->{$fd} and
+         croak "Epoll has already seen this filehandle; cannot add it a second time";
+      $self->{refaddr_for_fileno}->{$fd} = refaddr $handle;
+
+      if( defined $epoll->add( $handle, \@bits, $cb ) ) {
          # All OK
       }
       elsif( $! == EPERM ) {
@@ -279,11 +299,14 @@ sub watch_io
       $self->{masks}->{$fd} = $mask;
    }
    elsif( $mask != $curmask ) {
+      $self->{refaddr_for_fileno}->{$fd} == refaddr $handle or
+         croak "Epoll cannot cope with fd $fd changing handle under it";
+
       if( exists $fakeevents->{$fd} ) {
          $fakeevents->{$fd} = $mask;
       }
       else {
-         $epoll->modify( $handle, \@bits, $cb ) == 0
+         defined $epoll->modify( $handle, \@bits, $cb )
             or croak "Cannot EPOLL_CTL_MOD($fd,$mask) - $!";
       }
 
@@ -313,6 +336,9 @@ sub unwatch_io
 
    my $fakeevents = $self->{fakeevents};
 
+   $self->{refaddr_for_fileno}->{$fd} == refaddr $handle or
+      croak "Epoll cannot cope with fd $fd changing handle under it";
+
    if( $mask ) {
       if( exists $fakeevents->{$fd} ) {
          $fakeevents->{$fd} = $mask;
@@ -323,7 +349,7 @@ sub unwatch_io
          push @bits, 'out' if $mask & WATCH_WRITE;
          push @bits, 'hup' if $mask & WATCH_HUP;
 
-         $epoll->modify( $handle, \@bits, $cb ) == 0
+         defined $epoll->modify( $handle, \@bits, $cb )
             or croak "Cannot EPOLL_CTL_MOD($fd,$mask) - $!";
       }
 
@@ -334,12 +360,14 @@ sub unwatch_io
          delete $fakeevents->{$fd};
       }
       else {
-         $epoll->delete( $handle ) == 0
+         defined $epoll->delete( $handle )
             or croak "Cannot EPOLL_CTL_DEL($fd) - $!";
       }
 
       delete $self->{masks}->{$fd};
       delete $self->{callbacks}->{$fd};
+
+      delete $self->{refaddr_for_fileno}->{$fd};
    }
 }
 
@@ -355,8 +383,8 @@ sub watch_signal
    # Instead, we'll store a tiny piece of code that just sets a flag, and
    # check the flags on return from the epoll_pwait call.
 
-   $self->{signals}{$signal} = [ $code, 0, $SIG{$signal} ];
-   my $pending = \$self->{signals}{$signal}[1];
+   $self->{signals}{$signal} = SignalWatch( $code, 0, $SIG{$signal} );
+   my $pending = \$self->{signals}{$signal}->pending;
 
    my $signum = $self->signame2num( $signal );
    sigprocmask( SIG_BLOCK, POSIX::SigSet->new( $signum ) );
@@ -376,7 +404,7 @@ sub unwatch_signal
 
    # When we saved the original value, we might have got an undef. But %SIG
    # doesn't like having undef assigned back in, so we need to translate
-   $SIG{$signal} = $self->{signals}{$signal}[2] || 'DEFAULT';
+   $SIG{$signal} = $self->{signals}{$signal}->orig || 'DEFAULT';
 
    delete $self->{signals}{$signal};
    
